import aioacme

project = 'aioacme'
copyright = '2024, Timofei Kukushkin'
author = 'Timofei Kukushkin'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'myst_parser',
]
version = release = aioacme.__version__

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

html_theme = 'furo'
html_static_path = ['_static']

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'cryptography': ('https://cryptography.io/en/latest/', None),
}

autoclass_content = 'both'
autodoc_member_order = 'bysource'
autodoc_preserve_defaults = True



import asyncio
import uuid

import aiohttp
import pytest
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, rsa

import aioacme


@pytest.fixture(scope='session')
def event_loop():
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope='session')
async def pebble_url(docker_services, docker_ip) -> str:
    docker_services.start('pebble')
    port = docker_services.port_for('pebble', 14000)
    url = f'https://{docker_ip}:{port}'

    # wait for pebble to be ready
    async with aiohttp.ClientSession() as session:
        for _ in range(100):
            try:
                async with session.get(url, ssl=False):
                    break
            except aiohttp.ClientError as exc:
                last_exc = exc
                await asyncio.sleep(0.1)
        else:
            raise TimeoutError from last_exc

    yield url


@pytest.fixture()
async def client(pebble_url) -> aioacme.Client:
    account_key = ec.generate_private_key(ec.SECP256R1())
    async with aioacme.Client(directory_url=f'{pebble_url}/dir', ssl=False, account_key=account_key) as client:
        yield client


@pytest.fixture()
def domain():
    return f'{uuid.uuid1()}.example.com'


@pytest.fixture(scope='session')
def private_key():
    return rsa.generate_private_key(public_exponent=65537, key_size=2048)


@pytest.fixture()
def csr(domain, private_key):
    return (
        x509.CertificateSigningRequestBuilder()
        .subject_name(x509.Name([x509.NameAttribute(x509.NameOID.COMMON_NAME, domain)]))
        .add_extension(x509.SubjectAlternativeName([x509.DNSName(domain)]), critical=False)
        .sign(private_key, hashes.SHA256())
    )


@pytest.fixture(name='add_txt')
def add_txt_fixture(docker_services, docker_ip):
    docker_services.start('challtestsrv')
    port = docker_services.wait_for_service('challtestsrv', 8055)

    async def add_txt(domain: str, value: str) -> None:
        async with aiohttp.request(
            'POST', f'http://{docker_ip}:{port}/set-txt', json={'host': domain + '.', 'value': value}
        ) as response:
            response.raise_for_status()

    return add_txt



import asyncio
import hashlib
import sys
from collections.abc import AsyncIterator, Mapping, Sequence
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import datetime
from ssl import SSLContext
from types import TracebackType
from typing import Any, Final, Literal

import aiohttp
import orjson
import serpyco_rs
from cryptography import x509
from cryptography.hazmat.primitives import serialization

from aioacme._directories import LETS_ENCRYPT_STAGING_DIRECTORY
from aioacme._exceptions import AcmeError
from aioacme._jwk import JWK, jwk_thumbprint, make_jwk
from aioacme._jws import jws_encode
from aioacme._models import Account, AccountStatus, Authorization, Challenge, Error, Identifier, Order, RevocationReason
from aioacme._types import PrivateKeyTypes
from aioacme._utils import b64_encode

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

_account_serializer = serpyco_rs.Serializer(Account, camelcase_fields=True)
_identifier_serializer = serpyco_rs.Serializer(Identifier, camelcase_fields=True)
_error_serializer = serpyco_rs.Serializer(Error, camelcase_fields=True)
_authorization_serializer = serpyco_rs.Serializer(Authorization, camelcase_fields=True)
_order_serializer = serpyco_rs.Serializer(Order, camelcase_fields=True)
_challenge_serializer = serpyco_rs.Serializer(Challenge, camelcase_fields=True)


class Client:
    account_key: PrivateKeyTypes
    directory_url: Final[str]
    account_uri: str | None

    def __init__(
        self,
        *,
        account_key: PrivateKeyTypes,
        directory_url: str = LETS_ENCRYPT_STAGING_DIRECTORY,
        account_uri: str | None = None,
        ssl: SSLContext | bool = True,
    ) -> None:
        """
        Create new ACME client.

        By using this library you agree to the terms of service of the ACME server you are using.

        :param account_key: private key for account.
        :param directory_url: URL to get directory.
        :param account_uri: optional account URI, if not provided, it would be fetched on first request.
        :param ssl: SSL context.
        """
        self.account_key = account_key
        self.directory_url = directory_url
        self.account_uri = account_uri

        self._account_key_jwk = make_jwk(self.account_key)
        self._account_key_jwk_thumbprint = jwk_thumbprint(self._account_key_jwk)

        self._directory: _Directory | None = None
        self._session: aiohttp.ClientSession = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=ssl), cookie_jar=aiohttp.DummyCookieJar()
        )

        self._nonces: list[str] = []
        self._get_nonce_lock = asyncio.Lock()

    async def get_terms_of_service(self) -> str | None:
        return (await self._get_directory()).terms_of_service

    async def get_account(self) -> Account:
        """
        Get account info.

        Registers new account if it doesn't exist.
        """
        data: Any
        if self.account_uri is None:
            url = (await self._get_directory()).new_account
            data = {'termsOfServiceAgreed': True}
            jwk = self._account_key_jwk
        else:
            url = self.account_uri
            data = b''
            jwk = None

        async with self._request(url=url, data=data, jwk=jwk) as response:
            response_data = await response.json(loads=orjson.loads)

        account = _account_serializer.load({**response_data, 'uri': self.account_uri or response.headers['Location']})
        self.account_uri = account.uri
        return account

    async def update_account(
        self, *, contact: Sequence[str] | None = None, status: Literal[AccountStatus.deactivated] | None = None
    ) -> Account:
        """
        Update account info.

        :param contact: list of contacts.
        :param status: new account status, accepts only :attr:`.AccountStatus.deactivated`.
        """
        uri = await self._get_account_uri()
        data: dict[str, Any] = {}
        if contact is not None:
            data['contact'] = contact
        if status is not None:
            data['status'] = status.value
        async with self._request(url=uri, data=data) as response:
            response_data = await response.json(loads=orjson.loads)

        return _account_serializer.load({**response_data, 'uri': uri})

    async def new_order(
        self,
        identifiers: Sequence[Identifier],
        *,
        not_before: datetime | None = None,
        not_after: datetime | None = None,
    ) -> Order:
        """
        Create new certificate order.

        :param identifiers: list of identifiers (domains or ips).
        :param not_before: the requested value of the notBefore field in the certificate.
        :param not_after: the requested value of the notAfter field in the certificate.
        """
        url = (await self._get_directory()).new_order

        data: dict[str, Any] = {'identifiers': [_identifier_serializer.dump(i) for i in identifiers]}
        if not_before is not None:
            data['notBefore'] = not_before.isoformat()
        if not_after is not None:
            data['notAfter'] = not_after.isoformat()

        async with self._request(url, data=data) as response:
            response_data = await response.json(loads=orjson.loads)

        return _order_serializer.load({**response_data, 'uri': response.headers['Location']})

    async def get_order(self, order_uri: str) -> Order:
        """
        Get existing order.

        :param order_uri: order URI.
        """
        async with self._request(order_uri) as response:
            response_data = await response.json(loads=orjson.loads)
        return _order_serializer.load({**response_data, 'uri': order_uri})

    async def get_authorization(self, authorization_uri: str) -> Authorization:
        """
        Get existing authorization.

        :param authorization_uri: authorization URI.
        """
        async with self._request(authorization_uri) as response:
            response_data = await response.json(loads=orjson.loads)
        return _authorization_serializer.load({**response_data, 'uri': authorization_uri})

    def get_dns_challenge_domain(self, domain: str) -> str:
        """
        Generate domain for DNS challenge.

        :param domain: domain.
        """
        return f'_acme-challenge.{domain}'

    def get_dns_challenge_validation(self, token: str) -> str:
        """
        Generate TXT record value for DNS challenge.

        :param token: challenge token.
        """
        return b64_encode(
            hashlib.sha256(f'{token}.{self._account_key_jwk_thumbprint}'.encode('ascii')).digest()
        ).decode('ascii')

    async def answer_challenge(self, url: str) -> Challenge:
        """
        Respond to the challenge.

        https://datatracker.ietf.org/doc/html/rfc8555#section-7.5.1

        :param url: challenge url.
        """
        async with self._request(url, data={}) as response:
            response_data = await response.json(loads=orjson.loads)
        return _challenge_serializer.load(response_data)

    async def deactivate_authorization(self, uri: str) -> Authorization:
        """
        Deactivate authorization.

        https://datatracker.ietf.org/doc/html/rfc8555#section-7.5.2

        :param uri: authorization URI.
        """
        async with self._request(uri, data={'status': 'deactivated'}) as response:
            response_data = await response.json(loads=orjson.loads)
        return _authorization_serializer.load({**response_data, 'uri': uri})

    async def finalize_order(self, finalize: str, csr: x509.CertificateSigningRequest) -> Order:
        """
        Finalize the order by submitting the CSR to issue certificate.

        :param finalize: finalize uri.
        :param csr: CSR.
        """
        async with self._request(
            finalize,
            data={'csr': b64_encode(csr.public_bytes(serialization.Encoding.DER)).decode('ascii')},
        ) as response:
            response_data = await response.json(loads=orjson.loads)
        return _order_serializer.load({**response_data, 'uri': response.headers['Location']})

    async def get_certificate(self, certificate: str) -> bytes:
        """
        Download ready certificate in PEM format.

        :param certificate: certificate URL.
        """
        async with self._request(certificate) as response:
            return await response.read()

    async def revoke_certificate(
        self,
        certificate: x509.Certificate,
        key: PrivateKeyTypes | None = None,
        reason: RevocationReason = RevocationReason.unspecified,
    ) -> None:
        """
        Revoke certificate.

        https://datatracker.ietf.org/doc/html/rfc8555#section-7.6

        :param certificate: certificate.
        :param key: private key, if you want to revoke certificate without account key.
        :param reason: reason.
        """
        async with self._request(
            (await self._get_directory()).revoke_cert,
            data={
                'certificate': b64_encode(certificate.public_bytes(serialization.Encoding.DER)).decode('ascii'),
                'reason': reason.value,
            },
            jwk=make_jwk(key) if key else None,
            key=key,
        ):
            pass

    async def change_key(self, new_account_key: PrivateKeyTypes) -> None:
        """
        Change private key for current account.
        Further requests will be signed with this new key.

        https://datatracker.ietf.org/doc/html/rfc8555#section-7.6

        :param new_account_key: new private key.
        """
        new_account_key_jwk = make_jwk(new_account_key)

        url = (await self._get_directory()).key_change
        payload = await self._wrap_in_jws(
            url=url,
            data={
                'account': await self._get_account_uri(),
                'oldKey': self._account_key_jwk,
            },
            key=new_account_key,
            jwk=new_account_key_jwk,
            add_nonce=False,
        )
        async with self._request(url, data=payload):
            pass
        self.account_key = new_account_key
        self._account_key_jwk = new_account_key_jwk
        self._account_key_jwk_thumbprint = jwk_thumbprint(new_account_key_jwk)

    async def _get_account_uri(self) -> str:
        return self.account_uri or (await self.get_account()).uri

    @asynccontextmanager
    async def _request(
        self,
        url: str,
        *,
        data: Mapping[str, Any] | bytes = b'',
        jwk: JWK | None = None,
        key: PrivateKeyTypes | None = None,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        async with self._session.post(
            url,
            data=await self._wrap_in_jws(url=url, data=data, jwk=jwk, key=key),
            headers={'Content-Type': 'application/jose+json'},
        ) as response:
            self._add_nounce(response)

            if response.status < 300:
                yield response
                return

            try:
                error = _error_serializer.load(await response.json(loads=orjson.loads))
            except aiohttp.ContentTypeError:
                error = Error(type='unknown', detail='')

            if error.type != 'urn:ietf:params:acme:error:badNonce':
                raise AcmeError(error)

            # retry bad nonce
            async with self._request(url, data=data, jwk=jwk, key=key) as retried_response:
                yield retried_response

    async def _wrap_in_jws(  # noqa: PLR0913
        self,
        *,
        url: str,
        data: Mapping[str, Any] | bytes,
        jwk: JWK | None = None,
        key: PrivateKeyTypes | None = None,
        add_nonce: bool = True,
    ) -> bytes:
        headers: dict[str, Any] = {'url': url}
        if jwk:
            headers['jwk'] = jwk
        else:
            headers['kid'] = await self._get_account_uri()

        if add_nonce:
            headers['nonce'] = await self._get_nonce()

        payload = orjson.dumps(data) if not isinstance(data, bytes) else data
        return jws_encode(payload=payload, key=key or self.account_key, headers=headers)

    async def _get_nonce(self) -> str:
        async with self._get_nonce_lock:
            if not self._nonces:
                async with self._session.head((await self._get_directory()).new_nonce) as response:
                    self._nonces.append(response.headers['Replay-Nonce'])
            return self._nonces.pop()

    def _add_nounce(self, response: aiohttp.ClientResponse) -> None:
        self._nonces.append(response.headers['Replay-Nonce'])

    async def _get_directory(self) -> '_Directory':
        if self._directory is None:
            async with self._session.get(self.directory_url) as response:
                response_data = await response.json(loads=orjson.loads)

            self._directory = _Directory(
                new_account=response_data['newAccount'],
                new_nonce=response_data['newNonce'],
                new_order=response_data['newOrder'],
                revoke_cert=response_data['revokeCert'],
                key_change=response_data['keyChange'],
                terms_of_service=response_data.get('meta', {}).get('termsOfService'),
            )

        return self._directory

    async def close(self) -> None:
        await self._session.close()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc: BaseException | None, tb: TracebackType | None
    ) -> None:
        await asyncio.shield(self.close())


@dataclass(frozen=True, slots=True)
class _Directory:
    new_account: str
    new_nonce: str
    new_order: str
    revoke_cert: str
    key_change: str
    terms_of_service: str | None = None



import binascii
import sys
from collections.abc import Mapping
from typing import Any

import orjson
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, ed25519, rsa
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.asymmetric.utils import decode_dss_signature

from aioacme._types import PrivateKeyTypes
from aioacme._utils import b64_encode

if sys.version_info >= (3, 11):
    from typing import assert_never
else:
    from typing_extensions import assert_never


def jws_encode(
    payload: bytes,
    key: PrivateKeyTypes,
    headers: Mapping[str, Any],
) -> bytes:
    if isinstance(key, ec.EllipticCurvePrivateKey):
        alg = {ec.SECP256R1.name: 'ES256', ec.SECP384R1.name: 'ES384', ec.SECP521R1.name: 'ES512'}[key.curve.name]
    elif isinstance(key, rsa.RSAPrivateKey):
        alg = 'RS256'
    elif isinstance(key, ed25519.Ed25519PrivateKey):
        alg = 'EdDSA'
    else:
        assert_never(key)

    headers = {**headers, 'alg': alg}

    headers_b64 = b64_encode(orjson.dumps(headers))
    payload_b64 = b64_encode(payload)
    signing_input = headers_b64 + b'.' + payload_b64

    if isinstance(key, rsa.RSAPrivateKey):
        signature = key.sign(signing_input, PKCS1v15(), hashes.SHA256())
    elif isinstance(key, ec.EllipticCurvePrivateKey):
        hash_alg = {
            ec.SECP256R1.name: hashes.SHA256(),
            ec.SECP384R1.name: hashes.SHA384(),
            ec.SECP521R1.name: hashes.SHA512(),
        }[key.curve.name]
        signature = key.sign(signing_input, ec.ECDSA(hash_alg))
        signature = _der_to_raw_signature(signature, key.curve)
    elif isinstance(key, ed25519.Ed25519PrivateKey):
        signature = key.sign(signing_input)
    else:
        assert_never(key)

    return orjson.dumps(
        {
            'protected': headers_b64.decode('ascii'),
            'payload': payload_b64.decode('ascii'),
            'signature': b64_encode(signature).decode('ascii'),
        }
    )


def _der_to_raw_signature(der_sig: bytes, curve: ec.EllipticCurve) -> bytes:
    num_bytes = (curve.key_size + 7) // 8

    r, s = decode_dss_signature(der_sig)

    return _number_to_bytes(r, num_bytes) + _number_to_bytes(s, num_bytes)


def _number_to_bytes(num: int, num_bytes: int) -> bytes:
    padded_hex = '%0*x' % (2 * num_bytes, num)
    return binascii.a2b_hex(padded_hex.encode('ascii'))



from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Annotated

import dateutil.parser
from serpyco_rs.metadata import deserialize_with


class AccountStatus(Enum):
    """Account status."""

    valid = 'valid'
    """Account is active."""
    deactivated = 'deactivated'
    """Client requested deactivation of the account."""
    revoked = 'revoked'
    """Server revoked the account."""


@dataclass(frozen=True, slots=True, kw_only=True)
class Account:
    """Account object."""

    uri: str
    status: AccountStatus
    contact: Sequence[str] | None = None
    terms_of_service_agreed: bool | None = None
    orders: str | None = None
    """
    A URL from which a list of orders submitted by this account can be fetched
    """
    initial_ip: str | None = None
    created_at: Annotated[datetime, deserialize_with(dateutil.parser.isoparse)] | None = None


class OrderStatus(Enum):
    """Order status."""

    pending = 'pending'
    """The order is waiting for the client to satisfy authorizations challenges."""
    ready = 'ready'
    """All authorizations have been satisfied, the order is ready to finalize."""
    processing = 'processing'
    """The server is issuing the certificate."""
    valid = 'valid'
    """The certificate has been issued."""
    invalid = 'invalid'
    """The order has expired or one of its authorizations has failed."""


class AuthorizationStatus(Enum):
    """Authorization status."""

    pending = 'pending'
    """The authorization is waiting to be validated."""
    valid = 'valid'
    """The authorization has been successfully validated."""
    invalid = 'invalid'
    """The authorization has not been successfully validated."""
    deactivated = 'deactivated'
    """The client has deactivated the authorization."""
    expired = 'expired'
    """The authorization has expired."""
    revoked = 'revoked'
    """The server has revoked the authorization."""


class IdentifierType(Enum):
    """Identifier type."""

    dns = 'dns'
    ip = 'ip'


@dataclass(frozen=True, slots=True)
class Identifier:
    """Identifier object."""

    value: str
    type: IdentifierType = IdentifierType.dns


@dataclass(frozen=True, slots=True, kw_only=True)
class Error:
    """Error object."""

    type: str
    """A URI reference to a document with more information about the error."""
    detail: str
    """A short description of the error."""
    identifier: Identifier | None = None
    subproblems: Sequence['Error'] | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class Order:
    """Order object."""

    uri: str
    """URI"""
    identifiers: Sequence[Identifier]
    """An array of identifier objects that the order pertains to."""
    authorizations: Sequence[str]
    """
    For pending orders, the authorizations that the client needs to complete
    before the requested certificate can be issued,
    including unexpired authorizations that the client has completed in the
    past for identifiers specified in the order. The authorizations required
    are dictated by server policy; there may not be a 1:1 relationship between
    the order identifiers and the authorizations required.  For final orders
    (in the "valid" or "invalid" state), the authorizations that were
    completed. Each entry is a URL from which an authorization can be fetched.
    """
    status: OrderStatus
    """The status of this order."""
    expires: datetime | None = None
    """
    The timestamp after which the server will consider this order invalid.
    This field is REQUIRED for objects with "pending" or "valid"
    in the status field.
    """
    finalize: str
    """
    A URL that a CSR must be POSTed to once all of the order's authorizations
    are satisfied to finalize the order. The result of a successful
    finalization will be the population of the certificate URL for the order.
    """
    not_before: datetime | None = None
    """The requested value of the notBefore field in the certificate."""
    not_after: datetime | None = None
    """The requested value of the notAfter field in the certificate."""
    error: Error | None = None
    """The error that occurred while processing the order, if any."""
    certificate: str | None = None
    """
    A URL for the certificate that has been issued in response to this order.
    """


class ChallengeType(Enum):
    """Challenge type."""

    dns01 = 'dns-01'
    http01 = 'http-01'
    tlsalpn01 = 'tls-alpn-01'


class ChallengeStatus(Enum):
    """Challenge status."""

    pending = 'pending'
    """The challenge is waiting to be validated."""
    processing = 'processing'
    """The challenge is in the process of being validated."""
    valid = 'valid'
    """The challenge has been successfully validated."""
    invalid = 'invalid'
    """The challenge has not been successfully validated."""


@dataclass(frozen=True, slots=True, kw_only=True)
class Challenge:
    """Challenge object."""

    type: ChallengeType
    """The type of challenge encoded in the object."""
    url: str
    """The URL to which a response can be posted."""
    status: ChallengeStatus
    """The status of this challenge."""
    validated: datetime | None = None
    """
    The time at which the server validated
    this challenge, encoded in the format specified in [RFC3339].
    This field is REQUIRED if the "status" field is "valid".
    """
    error: Error | None = None
    """
    Error that occurred while the server was validating the challenge, if any.
    Multiple errors can be indicated by using subproblems Section 6.7.1.
    A challenge object with an error MUST have status equal to "invalid".
    """
    token: str
    """
    A random value that uniquely identifies the challenge. This value MUST
    have at least 128 bits of entropy. It MUST NOT contain any characters
    outside the base64url alphabet and MUST NOT include base64 padding
    characters ("=").
    """


@dataclass(frozen=True, slots=True, kw_only=True)
class Authorization:
    """Authorization object."""

    uri: str
    """URI"""
    identifier: Identifier
    """The identifier that the account is authorized to represent."""
    status: AuthorizationStatus
    """The status of this authorization."""
    expires: datetime | None = None
    """
    The timestamp after which the server
    will consider this authorization invalid, encoded in the format
    specified in [RFC3339].  This field is REQUIRED for objects with
    "valid" in the "status" field.
    """
    challenges: Sequence[Challenge]
    """
    For pending authorizations,
    the challenges that the client can fulfill in order to prove
    possession of the identifier.  For valid authorizations, the
    challenge that was validated.  For invalid authorizations, the
    challenge that was attempted and failed.  Each array entry is an
    object with parameters required to validate the challenge.  A
    client should attempt to fulfill one of these challenges, and a
    server should consider any one of the challenges sufficient to
    make the authorization valid.
    """
    wildcard: bool = False
    """
    This field MUST be present and true
    for authorizations created as a result of a newOrder request
    containing a DNS identifier with a value that was a wildcard
    domain name.  For other authorizations, it MUST be absent.
    """


class RevocationReason(Enum):
    """Reason to revoke certificate."""

    unspecified = 0
    key_compromise = 1
    ca_compromise = 2
    affiliation_changed = 3
    superseded = 4
    cessation_of_operation = 5
    certificate_hold = 6
    remove_from_crl = 8
    privilege_withdrawn = 9
    aa_compromise = 10



from aioacme._client import Client
from aioacme._directories import LETS_ENCRYPT_DIRECTORY, LETS_ENCRYPT_STAGING_DIRECTORY
from aioacme._exceptions import AcmeError
from aioacme._models import (
    Account,
    AccountStatus,
    Authorization,
    AuthorizationStatus,
    Challenge,
    ChallengeStatus,
    ChallengeType,
    Error,
    Identifier,
    IdentifierType,
    Order,
    OrderStatus,
    RevocationReason,
)
from aioacme._version import __version__, __version_tuple__

__all__ = [
    'Client',
    'LETS_ENCRYPT_DIRECTORY',
    'LETS_ENCRYPT_STAGING_DIRECTORY',
    'AcmeError',
    'Account',
    'AccountStatus',
    'Authorization',
    'AuthorizationStatus',
    'Challenge',
    'ChallengeStatus',
    'ChallengeType',
    'Error',
    'Identifier',
    'IdentifierType',
    'Order',
    'OrderStatus',
    'RevocationReason',
    '__version__',
    '__version_tuple__',
]



import asyncio
from datetime import datetime, timedelta, timezone
from unittest import mock

import pytest
from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import ec, ed25519, rsa

import aioacme


async def test_get_terms_of_service(client):
    # act
    result = await client.get_terms_of_service()

    # assert
    assert result == 'data:text/plain,Do%20what%20thou%20wilt'


async def test_get_account__not_registered(client):
    # act
    account = await client.get_account()

    # assert
    assert account == aioacme.Account(uri=mock.ANY, status=aioacme.AccountStatus.valid, orders=mock.ANY)


async def test_get_account__already_registered_but_uri_not_provided__return_existing(client):
    # arrange
    account = await client.get_account()

    # act
    async with aioacme.Client(
        account_key=client.account_key, directory_url=client.directory_url, ssl=False
    ) as new_client:
        result = await new_client.get_account()

    # assert
    assert result == account


async def test_get_account__already_registered_and_uri_provided__return_existing(client):
    # arrange
    account = await client.get_account()

    # act
    async with aioacme.Client(
        account_key=client.account_key, directory_url=client.directory_url, ssl=False, account_uri=account.uri
    ) as new_client:
        result = await new_client.get_account()

    # assert
    assert result == account


async def test_update_account__contant(client):
    # act
    account = await client.update_account(contact=['mailto:bar@example.com'])

    # assert
    assert account == aioacme.Account(
        uri=client.account_uri, status=aioacme.AccountStatus.valid, orders=mock.ANY, contact=['mailto:bar@example.com']
    )


async def test_update_account__deactivate(client):
    # act
    account = await client.update_account(status=aioacme.AccountStatus.deactivated)

    # assert
    assert account == aioacme.Account(uri=client.account_uri, status=aioacme.AccountStatus.deactivated, orders=mock.ANY)


async def test_new_order(client, domain, pebble_url):
    # arrange
    identifiers = [aioacme.Identifier(domain)]

    # act
    order = await client.new_order(identifiers)

    # assert
    assert order == aioacme.Order(
        uri=mock.ANY,
        identifiers=identifiers,
        authorizations=[mock.ANY],
        status=aioacme.OrderStatus.pending,
        expires=mock.ANY,
        finalize=mock.ANY,
    )
    assert order.uri.startswith(f'{pebble_url}/my-order')
    assert order.finalize.startswith(f'{pebble_url}/finalize-order')
    assert isinstance(order.expires, datetime)
    assert order.authorizations[0].startswith(f'{pebble_url}/authZ')


async def test_new_order__with_not_before_and_not_after(client, domain, pebble_url):
    # arrange
    identifiers = [aioacme.Identifier(domain)]
    not_before = datetime.now(timezone.utc)
    not_after = not_before + timedelta(days=30)

    # act
    order = await client.new_order(identifiers, not_before=not_before, not_after=not_after)

    # assert
    assert order == aioacme.Order(
        uri=mock.ANY,
        identifiers=identifiers,
        authorizations=[mock.ANY],
        status=aioacme.OrderStatus.pending,
        expires=mock.ANY,
        finalize=mock.ANY,
        not_before=not_before,
        not_after=not_after,
    )
    assert order.uri.startswith(f'{pebble_url}/my-order')
    assert order.finalize.startswith(f'{pebble_url}/finalize-order')
    assert isinstance(order.expires, datetime)
    assert order.authorizations[0].startswith(f'{pebble_url}/authZ')


async def test_get_order__no_order__exception(client, pebble_url):
    # act
    with pytest.raises(aioacme.AcmeError) as exc_info:
        await client.get_order(f'{pebble_url}/my-order/foo')

    assert exc_info.value.error == aioacme.Error(type='unknown', detail='')  # pebble doesn't return a proper error


async def test_get_order__order_exists__ok(client, domain):
    # arrange
    created_order = await client.new_order([aioacme.Identifier(domain)])

    # act
    result = await client.get_order(created_order.uri)

    # assert
    assert result == created_order


async def test_get_authorization(client, domain, pebble_url):
    # arrange
    identifier = aioacme.Identifier(domain)
    order = await client.new_order([identifier])
    authorization_uri = order.authorizations[0]

    # act
    result = await client.get_authorization(authorization_uri)

    # assert
    assert result == aioacme.Authorization(
        uri=authorization_uri,
        identifier=identifier,
        status=aioacme.AuthorizationStatus.pending,
        expires=mock.ANY,
        challenges=mock.ANY,
        wildcard=False,
    )
    assert isinstance(result.expires, datetime)

    challenges = sorted(result.challenges, key=lambda c: c.type.value)
    assert challenges == [
        aioacme.Challenge(
            type=aioacme.ChallengeType.dns01,
            url=mock.ANY,
            status=aioacme.ChallengeStatus.pending,
            token=mock.ANY,
        ),
        aioacme.Challenge(
            type=aioacme.ChallengeType.http01,
            url=mock.ANY,
            status=aioacme.ChallengeStatus.pending,
            token=mock.ANY,
        ),
        aioacme.Challenge(
            type=aioacme.ChallengeType.tlsalpn01,
            url=mock.ANY,
            status=aioacme.ChallengeStatus.pending,
            token=mock.ANY,
        ),
    ]
    assert all(c.url.startswith(pebble_url) for c in challenges)


async def test_finalize__pending_order__error(client, domain, csr):
    # arrange
    created_order = await client.new_order([aioacme.Identifier(domain)])

    # act
    with pytest.raises(aioacme.AcmeError) as exc_info:
        await client.finalize_order(created_order.finalize, csr)

    # assert
    assert exc_info.value.error == aioacme.Error(
        type='urn:ietf:params:acme:error:orderNotReady', detail='Order\'s status ("pending") was not ready'
    )


async def test_answer_challenge__invalid_record_value__challenge_error(client, domain, add_txt):
    order = await client.new_order([aioacme.Identifier(domain)])

    authorization = await client.get_authorization(order.authorizations[0])

    challenge = next(c for c in authorization.challenges if c.type == aioacme.ChallengeType.dns01)

    validation_domain = client.get_dns_challenge_domain(authorization.identifier.value)
    await add_txt(validation_domain, 'foo')

    challenge = await client.answer_challenge(challenge.url)
    assert challenge.status is aioacme.ChallengeStatus.processing

    await asyncio.sleep(0.5)

    authorization = await client.get_authorization(authorization.uri)
    assert authorization.status is aioacme.AuthorizationStatus.invalid

    challenge = next(c for c in authorization.challenges if c.type == aioacme.ChallengeType.dns01)
    assert challenge.status is aioacme.ChallengeStatus.invalid
    assert challenge.error == aioacme.Error(
        type='urn:ietf:params:acme:error:unauthorized', detail='Correct value not found for DNS challenge'
    )


async def test_deactivate_authorization(client, domain, add_txt):
    order = await client.new_order([aioacme.Identifier(domain)])

    authorization = await client.get_authorization(order.authorizations[0])

    challenge = next(c for c in authorization.challenges if c.type == aioacme.ChallengeType.dns01)

    validation_domain = client.get_dns_challenge_domain(authorization.identifier.value)
    validation = client.get_dns_challenge_validation(challenge.token)
    await add_txt(validation_domain, validation)

    challenge = await client.answer_challenge(challenge.url)
    assert challenge.status is aioacme.ChallengeStatus.processing

    await asyncio.sleep(0.5)

    authorization = await client.get_authorization(authorization.uri)
    assert authorization.status is aioacme.AuthorizationStatus.valid

    authorization = await client.deactivate_authorization(authorization.uri)
    assert authorization.status is aioacme.AuthorizationStatus.deactivated


async def test_integrational_ok(client, domain, csr, add_txt, private_key):
    order = await client.new_order([aioacme.Identifier(domain)])

    authorization = await client.get_authorization(order.authorizations[0])

    challenge = next(c for c in authorization.challenges if c.type == aioacme.ChallengeType.dns01)

    validation_domain = client.get_dns_challenge_domain(authorization.identifier.value)
    validation = client.get_dns_challenge_validation(challenge.token)
    await add_txt(validation_domain, validation)

    challenge = await client.answer_challenge(challenge.url)
    assert challenge.status is aioacme.ChallengeStatus.processing

    await asyncio.sleep(0.5)

    authorization = await client.get_authorization(authorization.uri)
    assert authorization.status is aioacme.AuthorizationStatus.valid
    assert authorization.challenges == [
        aioacme.Challenge(
            type=aioacme.ChallengeType.dns01,
            url=challenge.url,
            status=aioacme.ChallengeStatus.valid,
            validated=mock.ANY,
            token=challenge.token,
        )
    ]
    assert isinstance(authorization.challenges[0].validated, datetime)

    order = await client.get_order(order.uri)
    assert order.status is aioacme.OrderStatus.ready

    order = await client.finalize_order(order.finalize, csr)
    assert order.status is aioacme.OrderStatus.processing

    await asyncio.sleep(0.5)
    order = await client.get_order(order.uri)
    assert order.status is aioacme.OrderStatus.valid

    certificate = await client.get_certificate(order.certificate)

    assert certificate.startswith(b'---')

    await client.revoke_certificate(x509.load_pem_x509_certificate(certificate), private_key)


@pytest.mark.parametrize(
    'get_new_key',
    [
        pytest.param(lambda: rsa.generate_private_key(public_exponent=65537, key_size=2048), id='RSA_2048'),
        pytest.param(lambda: rsa.generate_private_key(public_exponent=65537, key_size=4096), id='RSA_4096'),
        pytest.param(lambda: ec.generate_private_key(ec.SECP256R1()), id='SECP256R1'),
        pytest.param(lambda: ec.generate_private_key(ec.SECP384R1()), id='SECP384R1'),
        pytest.param(lambda: ec.generate_private_key(ec.SECP521R1()), id='SECP521R1'),
        pytest.param(
            ed25519.Ed25519PrivateKey.generate,
            marks=[pytest.mark.xfail(reason='Ed25519 is not supported')],
            id='Ed25519',
        ),
    ],
)
async def test_change_key(client, domain, get_new_key):
    # arrange
    new_key = get_new_key()

    # act
    await client.change_key(new_key)

    # assert
    await client.new_order([aioacme.Identifier(domain)])



import hashlib
import sys
from collections.abc import Mapping
from typing import Any, NewType

import orjson
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec, ed25519, rsa

from aioacme._types import PrivateKeyTypes
from aioacme._utils import b64_encode

if sys.version_info >= (3, 11):
    from typing import assert_never
else:
    from typing_extensions import assert_never

JWK = NewType('JWK', Mapping[str, Any])


def make_jwk(key: PrivateKeyTypes) -> JWK:
    if isinstance(key, ec.EllipticCurvePrivateKey):
        ec_public_key = key.public_key()
        ec_public_numbers = ec_public_key.public_numbers()

        crv = {'secp256r1': 'P-256', 'secp384r1': 'P-384', 'secp521r1': 'P-521'}[ec_public_key.curve.name]
        byte_length = (ec_public_key.curve.key_size + 7) // 8

        return JWK(
            {
                'kty': 'EC',
                'crv': crv,
                'x': _int_to_base64url(ec_public_numbers.x, byte_length=byte_length),
                'y': _int_to_base64url(ec_public_numbers.y, byte_length=byte_length),
            }
        )

    if isinstance(key, rsa.RSAPrivateKey):
        rsa_public_numbers = key.public_key().public_numbers()
        return JWK(
            {
                'kty': 'RSA',
                'e': _int_to_base64url(rsa_public_numbers.e),
                'n': _int_to_base64url(rsa_public_numbers.n),
            }
        )

    if isinstance(key, ed25519.Ed25519PrivateKey):
        return JWK(
            {
                'kty': 'OKP',
                'crv': 'Ed25519',
                'x': b64_encode(
                    key.public_key().public_bytes(serialization.Encoding.Raw, serialization.PublicFormat.Raw)
                ).decode('ascii'),
            }
        )

    assert_never(key)


def jwk_thumbprint(jwk: JWK) -> str:
    jwk_str = orjson.dumps(jwk, option=orjson.OPT_SORT_KEYS)
    jwk_hash = hashlib.sha256(jwk_str).digest()
    return b64_encode(jwk_hash).decode('ascii')


def _int_to_base64url(number: int, *, byte_length: int | None = None) -> str:
    byte_length = byte_length or (number.bit_length() + 7) // 8
    return b64_encode(number.to_bytes(byte_length, byteorder='big')).decode('ascii')
